--- mhddfs-0.1.38.orig/src/main.c	2011-05-04 04:29:42.000000000 -0400
+++ mhddfs-0.1.38.new/src/main.c	2012-06-17 14:39:05.425433236 -0400
@@ -298,7 +298,7 @@
 
 	mhdd_debug(MHDD_INFO, "mhdd_internal_open: new file %s\n", file);
 
-	if ((dir_id = get_free_dir()) < 0) {
+	if ((dir_id = get_parent_dir(file)) < 0) {
 		errno = ENOSPC;
 		return -errno;
 	}
--- mhddfs-0.1.38.orig/src/tools.c	2011-05-04 04:29:42.000000000 -0400
+++ mhddfs-0.1.38.new/src/tools.c	2012-06-17 14:38:56.245097852 -0400
@@ -421,6 +421,84 @@
 	return(path);
 }
 
+/*
+ * This is used by get_free_parent_dir() to verify
+ * that an existing subdir has enough space for another file.
+ * Code was cloned/hacked from get_free_dir(),
+ *
+ * Here we are willing to relax storage limits by a factor of 2
+ * if it helps keep files together.
+ */
+fsblkcnt_t get_parent_bavail (int dir_id)
+{
+	struct statvfs stf;
+	if (statvfs(mhdd.dirs[dir_id], &stf) != 0)
+		return 0;
+
+	off_t threshold, move_limit = mhdd.move_limit;
+	if (move_limit <= 100)
+		threshold = 10;				/* minimum percent-free */
+	else
+		threshold = 200ll * 1024 * 1024 * 1024;	/* minimum space free */
+	if (move_limit > threshold)
+		move_limit /= 2;
+	else if (move_limit > (threshold/2))
+		move_limit = (threshold/2);
+
+	if (move_limit < 100) {
+		/* minimum percent-free */
+		fsblkcnt_t perclimit = stf.f_blocks;
+		perclimit *= move_limit + 1;
+		perclimit /= 100;
+		if (stf.f_bavail >= perclimit)
+			return stf.f_bavail;
+	} else {
+		/* minimum space free */
+		fsblkcnt_t space = stf.f_bsize;
+		space *= stf.f_bavail;
+		if (space >= move_limit)
+			return stf.f_bavail;
+	}
+	return 0;
+}
+
+/*
+ * We have hacked "mhdd_internal_open(CREATE_FUNCTION)" to call here
+ * instead of get_free_dir() when deciding where to create a *new* file.
+ *
+ * Here we try to place the new file in an existing instance
+ * of the same subdir if at all possible, before falling back
+ * to the original regular get_free_dir() function for placement.
+ *
+ * The idea is to co-locate files from the same directory
+ * together onto the same underlying filesystem as much as possible.
+ */
+int get_parent_dir(const char *file)
+{
+	fsblkcnt_t best_bavail = 0;
+	int i, best = -1;
+
+	for (i=0; i<mhdd.cdirs; i++)
+	{
+		char *path=create_path(mhdd.dirs[i], file);
+		if (!path) break;
+		char *slash = strrchr(path, '/');
+		if (slash && slash != path) {
+			struct stat st;
+			*slash = '\0';
+			if (lstat(path, &st)==0) {
+				fsblkcnt_t bavail = get_parent_bavail(i);
+				if (bavail > best_bavail) {
+					best_bavail = bavail;
+					best = i;
+				}
+			}
+		}
+		free(path);
+	}
+	return best ? best : get_free_dir();
+}
+
 char * find_path(const char *file)
 {
 	int i;
--- mhddfs-0.1.38.orig/src/tools.h	2011-05-04 04:29:42.000000000 -0400
+++ mhddfs-0.1.38.new/src/tools.h	2012-06-17 14:38:56.245097852 -0400
@@ -26,6 +26,7 @@
 
 #include "flist.h"
 
+int get_parent_dir(const char *file);
 int get_free_dir(void);
 char * create_path(const char *dir, const char * file);
 char * find_path(const char *file);
